#include <stdio.h>
#include <stdlib.h>
#include "second.h"

void
report (int nx, int lx, float op, float tm)
{
  float mf = -1.;
  if (tm > 0.)
    mf = op / tm;
  printf ("%10d%10d%10.2f%10.2f\n", nx, lx, tm, mf);
}

extern void  svm   (float, float *, float *, int);
extern void  svmva (float, float *, float *, float *, int);
extern void  svmsa (float, float *, float, float *, int);
extern void  vvm   (float *, float *, float *, int);
extern void  vva   (float *, float *, float *, int);
extern float vsum  (float *, int);
extern float vdot  (float *, float *, int);

void
main (int argc, char *argv[])
{
#define LOOP 2000000
#define NM 7
#define NNX 500
  int OP, i, j, k, nx, nnx, lx, mmx;
  float a = .1, d, *x, *y, op, tm;
  long str, LX, NX[NM] =
  {500, 1000, 5000, 10000, 50000, 100000, 500000};
  if (argc < 2)
    {
      printf ("Usage : %s loop-multiplier [stride]\n", argv[0]);
      exit (1);
    }
  sscanf (argv[1], "%ld", &LX);
  str = NNX;
  OP = 10;
  if (argc > 2)
    sscanf (argv[2], "%ld", &str);
  if (str <= 0)
    str = NX[NM];
  LX *= LOOP;
  printf ("      Size     Loops      Time    MFlops\n");
  for (j = 0; j < NM; j++)
    {
      nx = NX[j];
      lx = LX / nx;
#ifdef HELP_GCC
      x = (float *) malloc (2 * nx * sizeof (float));
      y = x + nx;
#else
      x = (float *) malloc (nx * sizeof (float));
      y = (float *) malloc (nx * sizeof (float));
#endif
      if (x == (float *) 0)
	{
	  puts ("Cannot allocate memory for x[]\n");
	  exit (-1);
	}
      if (y == (float *) 0)
	{
	  puts ("Cannot allocate memory for y[]\n");
	  exit (-1);
	}
      for (i = 0; i < nx; i++)
	{
	  x[i] = 1.;
	  y[i] = 0.;
	}
      op = (1.e-3 * (OP * nx)) * (1.e-3 * lx);
      d = 0.;
      tm = second ();
      nnx = (nx < str ? nx : str);
      for (k = 0; k < lx; k++)
	for (i = 0; i < nx; i += nnx)
	  {
	    mmx = nx - i;
	    mmx = (mmx < nnx ? mmx : nnx);
	    svm (a, &x[i], &y[i], mmx);
	    svmva (a, &x[i], &y[i], &y[i], mmx);
	    svmsa (a, &x[i], a, &y[i], mmx);
	    vvm (&x[i], &y[i], &y[i], mmx);
	    vva (&x[i], &y[i], &y[i], mmx);
	    d += vsum (&x[i], mmx);
	    d -= vdot (&x[i], &x[i], mmx);
	  }
      tm = second () - tm;
      report (nx, lx, op, tm);
      free (x);
#ifndef HELP_GCC
      free (y);
#endif
    }
  exit(0);
}
void
svm (float a, float *x, float *y, int n)
{
  int i, m;
  float t1, t2;
  m = n & ~7L;
  for (i = 0; i < m; i += 8)
    {
#ifdef HELP_GCC
      float t3, t4, t5, t6, t7, t8;

      t1 = a * x[i + 0];
      t2 = a * x[i + 1];
      t3 = a * x[i + 2];
      t4 = a * x[i + 3];
      t5 = a * x[i + 4];
      t6 = a * x[i + 5];
      t7 = a * x[i + 6];
      t8 = a * x[i + 7];

      y[i + 0] = t1;
      y[i + 1] = t2;
      y[i + 2] = t3;
      y[i + 3] = t4;
      y[i + 4] = t5;
      y[i + 5] = t6;
      y[i + 6] = t7;
      y[i + 7] = t8;
#else
      t1 = a * x[i + 0];
      t2 = a * x[i + 1];
      y[i + 0] = t1;
      t1 = a * x[i + 2];
      y[i + 1] = t2;
      t2 = a * x[i + 3];
      y[i + 2] = t1;
      t1 = a * x[i + 4];
      y[i + 3] = t2;
      t2 = a * x[i + 5];
      y[i + 4] = t1;
      t1 = a * x[i + 6];
      y[i + 5] = t2;
      t2 = a * x[i + 7];
      y[i + 6] = t1;
      y[i + 7] = t2;
#endif
    }
  for (i = m; i < n; i++)
    y[i] = a * x[i];
  return;
}
void
svmva (float a, float *x, float *y, float *z, int n)
{
  int i, m;
  float t1, t2;
  m = n & ~7L;
  for (i = 0; i < m; i += 8)
    {
#ifdef HELP_GCC
      float t3, t4, t5, t6, t7, t8;

      t1 = a * x[i + 0];
      t2 = a * x[i + 1];
      t1 += y[i + 0];
      t2 += y[i + 1];
      t3 = a * x[i + 2];
      t3 += y[i + 2];
      t6 = a * x[i + 3];
      t6 += y[i + 3];
      t4 = a * x[i + 4];
      t4 += y[i + 4];
      t7 = a * x[i + 5];
      t7 += y[i + 5];
      t5 = a * x[i + 6];
      t8 = a * x[i + 7];
      t5 += y[i + 6];
      t8 += y[i + 7];

      z[i + 0] = t1;
      z[i + 1] = t2;
      z[i + 2] = t3;
      z[i + 3] = t6;
      z[i + 4] = t4;
      z[i + 5] = t7;
      z[i + 6] = t5;
      z[i + 7] = t8;
#else
      t1 = a * x[i + 0];
      t2 = a * x[i + 1];
      t1 += y[i + 0];
      t2 += y[i + 1];
      z[i + 0] = t1;
      t1 = a * x[i + 2];
      z[i + 1] = t2;
      t1 += y[i + 2];
      t2 = a * x[i + 3];
      z[i + 2] = t1;
      t2 += y[i + 3];
      t1 = a * x[i + 4];
      z[i + 3] = t2;
      t1 += y[i + 4];
      t2 = a * x[i + 5];
      z[i + 4] = t1;
      t2 += y[i + 5];
      t1 = a * x[i + 6];
      z[i + 5] = t2;
      t2 = a * x[i + 7];
      t1 += y[i + 6];
      t2 += y[i + 7];
      z[i + 6] = t1;
      z[i + 7] = t2;
#endif
    }
  for (i = m; i < n; i++)
    z[i] = y[i] + a * x[i];
  return;
}
void
svmsa (float a, float *x, float b, float *y, int n)
{
  int i, m;
  float t1, t2;
  m = n & ~7L;
  for (i = 0; i < m; i += 8)
    {
#ifdef HELP_GCC
      float t3, t4, t5, t6, t7, t8;

      t1 = a * x[i + 0];
      t2 = a * x[i + 1];
      t1 += b;
      t2 += b;
      t3 = a * x[i + 2];
      t3 += b;
      t6 = a * x[i + 3];
      t6 += b;
      t4 = a * x[i + 4];
      t4 += b;
      t7 = a * x[i + 5];
      t7 += b;
      t5 = a * x[i + 6];
      t8 = a * x[i + 7];
      t5 += b;
      t8 += b;

      y[i + 0] = t1;
      y[i + 1] = t2;
      y[i + 2] = t3;
      y[i + 3] = t6;
      y[i + 4] = t4;
      y[i + 5] = t7;
      y[i + 6] = t5;
      y[i + 7] = t8;
#else
      t1 = a * x[i + 0];
      t2 = a * x[i + 1];
      t1 += b;
      t2 += b;
      y[i + 0] = t1;
      t1 = a * x[i + 2];
      y[i + 1] = t2;
      t1 += b;
      t2 = a * x[i + 3];
      y[i + 2] = t1;
      t2 += b;
      t1 = a * x[i + 4];
      y[i + 3] = t2;
      t1 += b;
      t2 = a * x[i + 5];
      y[i + 4] = t1;
      t2 += b;
      t1 = a * x[i + 6];
      y[i + 5] = t2;
      t2 = a * x[i + 7];
      t1 += b;
      t2 += b;
      y[i + 6] = t1;
      y[i + 7] = t2;
#endif
    }
  for (i = m; i < n; i++)
    y[i] = a * x[i] + b;
  return;
}
void
vvm (float *x, float *y, float *z, int n)
{
  int i, m;
  float t1, t2;
  m = n & ~7L;
  for (i = 0; i < m; i += 8)
    {
#ifdef HELP_GCC
      float t3, t4, t5, t6, t7, t8;

      t1 = x[i + 0] * y[i + 0];
      t2 = x[i + 1] * y[i + 1];
      t3 = x[i + 2] * y[i + 2];
      t4 = x[i + 3] * y[i + 3];
      t5 = x[i + 4] * y[i + 4];
      t6 = x[i + 5] * y[i + 5];
      t7 = x[i + 6] * y[i + 6];
      t8 = x[i + 7] * y[i + 7];

      z[i + 0] = t1;
      z[i + 1] = t2;
      z[i + 2] = t3;
      z[i + 3] = t4;
      z[i + 4] = t5;
      z[i + 5] = t6;
      z[i + 6] = t7;
      z[i + 7] = t8;
#else
      t1 = x[i + 0] * y[i + 0];
      t2 = x[i + 1] * y[i + 1];
      z[i + 0] = t1;
      t1 = x[i + 2] * y[i + 2];
      z[i + 1] = t2;
      t2 = x[i + 3] * y[i + 3];
      z[i + 2] = t1;
      t1 = x[i + 4] * y[i + 4];
      z[i + 3] = t2;
      t2 = x[i + 5] * y[i + 5];
      z[i + 4] = t1;
      t1 = x[i + 6] * y[i + 6];
      z[i + 5] = t2;
      t2 = x[i + 7] * y[i + 7];
      z[i + 6] = t1;
      z[i + 7] = t2;
#endif
    }
  for (i = m; i < n; i++)
    z[i] = x[i] * y[i];
  return;
}
void
vva (float *x, float *y, float *z, int n)
{
  int i, m;
  float t1, t2;
  m = n & ~7L;
  for (i = 0; i < m; i += 8)
    {
#ifdef HELP_GCC
      float t3, t4, t5, t6, t7, t8;

      t1 = x[i + 0] + y[i + 0];
      t2 = x[i + 1] + y[i + 1];
      t3 = x[i + 2] + y[i + 2];
      t4 = x[i + 3] + y[i + 3];
      t5 = x[i + 4] + y[i + 4];
      t6 = x[i + 5] + y[i + 5];
      t7 = x[i + 6] + y[i + 6];
      t8 = x[i + 7] + y[i + 7];

      z[i + 0] = t1;
      z[i + 1] = t2;
      z[i + 2] = t3;
      z[i + 3] = t4;
      z[i + 4] = t5;
      z[i + 5] = t6;
      z[i + 6] = t7;
      z[i + 7] = t8;
#else
      t1 = x[i + 0] + y[i + 0];
      t2 = x[i + 1] + y[i + 1];
      z[i + 0] = t1;
      t1 = x[i + 2] + y[i + 2];
      z[i + 1] = t2;
      t2 = x[i + 3] + y[i + 3];
      z[i + 2] = t1;
      t1 = x[i + 4] + y[i + 4];
      z[i + 3] = t2;
      t2 = x[i + 5] + y[i + 5];
      z[i + 4] = t1;
      t1 = x[i + 6] + y[i + 6];
      z[i + 5] = t2;
      t2 = x[i + 7] + y[i + 7];
      z[i + 6] = t1;
      z[i + 7] = t2;
#endif
    }
  for (i = m; i < n; i++)
    z[i] = x[i] + y[i];
  return;
}
float
vsum (float *x, int n)
{
  int i, m;
  float vs, t1, t2;
  m = n & ~7L;
  t1 = t2 = 0.f;
#ifdef HELP_GCC
  {
    float t3 = 0.0, t4 = 0.0, t5 = 0.0, t6 = 0.0 , t7 = 0.0, t8 = 0.0;

    for (i = 0; i < m; i += 8)
      {
	t1 += x[i + 0];
	t2 += x[i + 1];
	t3 += x[i + 4];
	t4 += x[i + 5];
	t5 += x[i + 2];
	t6 += x[i + 3];
	t7 += x[i + 6];
	t8 += x[i + 7];
      }
    vs = t1 + t2 + t3 + t4 + t5 + t6 + t7 + t8;
  }
#else
  for (i = 0; i < m; i += 8)
    {
      t1 += x[i + 0];
      t2 += x[i + 1];
      t1 += x[i + 4];
      t2 += x[i + 5];
      t1 += x[i + 2];
      t2 += x[i + 3];
      t1 += x[i + 6];
      t2 += x[i + 7];
    }
  vs = t1 + t2;
#endif
  for (i = m; i < n; i++)
    vs += x[i];
  return vs;
}
float
vdot (float *x, float *y, int n)
{
  int i, m;
  float vs, t1, t2, s1, s2;
  m = n & ~7L;
  s1 = s2 = t1 = t2 = 0.f;
  for (i = 0; i < m; i += 8)
    {
#ifdef HELP_GCC
      float x0, x1, x2, x3, x4, x5, x6, x7;
      float y0, y1, y2, y3, y4, y5, y6, y7;

      x0 = x[i + 0]; x1 = x[i + 1]; x2 = x[i + 2]; x3 = x[i + 3];
      x4 = x[i + 4]; x5 = x[i + 5]; x6 = x[i + 6]; x7 = x[i + 7];
      y0 = y[i + 0]; y1 = y[i + 1]; y2 = y[i + 2]; y3 = y[i + 3];
      y4 = y[i + 4]; y5 = y[i + 5]; y6 = y[i + 6]; y7 = y[i + 7];

      t1 = x0 * y0;
      s2 += t2;
      t2 = x1 * y1;
      s1 += t1;
      t1 = x2 * y2;
      s2 += t2;
      t2 = x3 * y3;
      s1 += t1;
      t1 = x4 * y4;
      s2 += t2;
      t2 = x5 * y5;
      s1 += t1;
      t1 = x6 * y6;
      s2 += t2;
      t2 = x7 * y7;
      s1 += t1;
#else
      t1 = x[i + 0] * y[i + 0];
      s2 += t2;
      t2 = x[i + 1] * y[i + 1];
      s1 += t1;
      t1 = x[i + 2] * y[i + 2];
      s2 += t2;
      t2 = x[i + 3] * y[i + 3];
      s1 += t1;
      t1 = x[i + 4] * y[i + 4];
      s2 += t2;
      t2 = x[i + 5] * y[i + 5];
      s1 += t1;
      t1 = x[i + 6] * y[i + 6];
      s2 += t2;
      t2 = x[i + 7] * y[i + 7];
      s1 += t1;
#endif
    }
  vs = s1 + s2;
  for (i = m; i < n; i++)
    vs += x[i] * y[i];
  return vs;
}
